from zope.interface import implements
from zope.component import queryMultiAdapter, getUtility, adapts
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.interfaces import ISiteRoot
from Products.GenericSetup.utils import XMLAdapterBase
from Products.GenericSetup.interfaces import IBody, ISetupEnviron
from p4a.subtyper.interfaces import ISubtyped, ISubtyper


class SubtyperXMLAdapter(XMLAdapterBase):
    """A GenericSetup XMLAdapter the im- and exports subtyped object
    metadata.
    """
    implements(IBody)
    # XXX doesn't necessarily need to adapt ISiteRoot does it?
    adapts(ISiteRoot, ISetupEnviron)
    
    _LOGGER_ID = name = filename = 'subtyper'
    
    def _exportNode(self):
        node = self._doc.createElement('subtyper')
        subtyper = getUtility(ISubtyper)
        pc = getToolByName(self.context, 'portal_catalog')
        subtyped_items = pc(object_provides=ISubtyped.__identifier__)
        for subtyped_item_brain in subtyped_items:
            child = self._doc.createElement('subtype')
            subtyped_item = subtyped_item_brain.getObject()
            descriptor = subtyper.existing_type(subtyped_item)
            child.setAttribute('descriptor', descriptor.name)
            obj_path = "/" + "/".join(subtyped_item.getPhysicalPath()[2:])
            child.setAttribute('path', obj_path)
            node.appendChild(child)
        
        self._logger.info('Subtyper exported')
        return node
    
    def _importNode(self, node):
        # XXX do something about env purge?
        self.subtyper = getUtility(ISubtyper)
        for child in node.childNodes:
            if child.nodeName.lower() == 'subtype':
                self._initSubtypeNode(child)
        
        self._logger.info('Subtyper imported')
    
    def _initSubtypeNode(self, node):
        descriptor = node.getAttribute('descriptor')
        # path is required
        # XXX should this raise the error or just log it...
        if not node.hasAttribute('path'):
            raise KeyError, "A 'path' is required when importing subtype settings"
        path = node.getAttribute('path').encode()
        remove = node.hasAttribute('remove') and \
                 self._convertToBoolean(node.getAttribute('remove'))
        # remove all subtyping or subtypes of a specific descriptor
        if remove and path == "*":
            pc = getToolByName(self.context, 'portal_catalog')
            subtyped_items = pc(object_provides=ISubtyped.__identifier__)
            for subtyped_item_brain in subtyped_items:
                subtyped_item = subtyped_item_brain.getObject()
                if descriptor == "*":
                    # remove all subtypes on this objects
                    self._removeSubtype(subtyped_item)
                else:
                    self._removeSubtype(subtyped_item, descriptor)
        else:
            path = path[1:]
            obj = self.context.restrictedTraverse(path, None)
            # XXX should we just log this ?
            if obj is None:
                # XXX should this raise the error or just log it...
                raise KeyError, "Cannot find object at path %s" % path
            if remove:
                if descriptor == "*":
                    # remove all subtypes on this objects
                    self._removeSubtype(obj)
                else:
                    self._removeSubtype(obj, descriptor)
            else:
                self._changeSubtype(obj, descriptor)
    
    def _removeSubtype(self, obj, descriptor=None):
        if descriptor is not None:
            current_subtype = self.subtyper.existing_type(obj)
            # XXX logging? errors here?
            if current_subtype is not None \
              and current_subtype.name == descriptor:
                obj_path = "/".join(obj.getPhysicalPath())
                msg = "'%s' removed from %s" % (descriptor, obj_path)
                self._logger.info(msg)
                self.subtyper.remove_type(obj)
        else:
            msg = "removed subtype from %s" % obj_path
            self._logger.info(msg)
            self.subtyper.remove_type(obj)
    
    def _changeSubtype(self, obj, descriptor):
        possible_descriptors = [
            possibility.name
            for possibility in self.subtyper.possible_types(obj)
        ]
        current_subtype_name = ''
        current_subtype = self.subtyper.existing_type(obj)
        if current_subtype is not None:
            current_subtype_name = current_subtype.name
        # make sure the descriptor isn't already set and that it
        # is a possible type to switch to
        if descriptor != current_subtype_name \
          and descriptor in possible_descriptors:
            obj_path = "/".join(obj.getPhysicalPath())
            msg = "'%s' set on %s" % (descriptor, obj_path)
            self._logger.info(msg)
            self.subtyper.change_type(obj, descriptor)

def importSubtyper(context):
    """Import subtyper settings
    """
    site = context.getSite()
    importer = queryMultiAdapter((site, context), IBody, name=u'p4a.subtyper')
    if importer:
        filename = '%s%s' % (importer.name, importer.suffix)
        body = context.readDataFile(filename)
        if body is not None:
            importer.filename = filename # for error reporting
            importer.body = body

def exportSubtyper(context):
    """Export subtyper settings
    """
    site = context.getSite()
    exporter = queryMultiAdapter((site, context), IBody, name=u'p4a.subtyper')
    if exporter:
        filename = '%s%s' % (exporter.name, exporter.suffix)
        body = exporter.body
        if body is not None:
            context.writeDataFile(filename, body, exporter.mime_type)
